import numpy as np
cimport numpy as np
from numpy cimport int_t
cimport cython
from cython.parallel import prange
#@cython.boundscheck(False)
@cython.wraparound(False)

def greedy(np.ndarray[int_t, mode='c'] Is, np.ndarray[int_t, mode='c'] Js, np.ndarray[int_t, mode='c'] parities, np.ndarray[int_t, mode='c'] flips):
    """Performs greedy algorithm.
    
    The input represents a graph where ideally edge (Is[k], Js[k]) would have
    parity parities[k].  That is, ideally
    
    (flips[Is[k]] + flips[Js[k]] % 2)  == parites[k]
    
    These constraints are satisfied one by one in a greedy fashion.
    The python variable prog is incrmeented for each constraint satisfied.
    """
    
    cdef int_t i, j
    cdef int_t n_constraints = len(Is)
    cdef int_t sI, sJ
    cdef int_t flipJs 
    cdef int_t I, J
    cdef int_t parity
    cdef int_t n_variables = len(flips)
    cdef np.ndarray[np.int_t] parents = np.arange(n_variables)
    cdef int cur, parent, child
    
    # Set up a tree structure.
    adj = []
    for i in range(n_variables):
        adj.append([])
    for i in range(n_constraints):
        I = Is[i]
        J = Js[i]
        parity = parities[i]
        
        # Find parents and compress paths.
        sI = I
        while parents[sI] != sI:
            sI = parents[sI]
        parents[I] = sI
        sJ = J
        while parents[sJ] != sJ:
            sJ = parents[sJ]
        parents[J] = sJ
        if sI != sJ:
            # Save the edge.    
            adj[I].append((J, parity))
            adj[J].append((I, parity))
            # Join these.
            parents[sJ] = sI
    
    for cur in np.nonzero(parents==np.arange(n_variables))[0]:    
        parent = -1
        stack = [(cur, parent)]
        while len(stack) > 0:
            cur, parent = stack.pop()
            for (child, parity) in adj[cur]:
                if child != parent:
                    
                    if parent != -1:
                        # Set the child flip.
                        flips[child] = (flips[cur] + parity) %2
                    
                    # And add it to the stack.
                    stack.append((child, cur))